@@ -2,12 +2,12 @@
 // Program to pull the information out of various types of EXIF digital 
 // camera files and show it in a reasonably consistent way
 //
-// Version 2.84
+// Version 2.86
 //
 // Compiling under Windows:  
 //   Make sure you have Microsoft's compiler on the path, then run make.bat
 //
-// Dec 1999 - Oct 2008
+// Dec 1999 - Feb 2009
 //
 // by Matthias Wandel   www.sentex.net/~mwandel
 //--------------------------------------------------------------------------
@@ -15,7 +15,7 @@
 
 #include <sys/stat.h>
 
-#define JHEAD_VERSION "2.84"
+#define JHEAD_VERSION "2.86"
 
 // This #define turns on features that are too very specific to 
 // how I organize my photos.  Best to ignore everything inside #ifdef MATTHIAS
@@ -35,8 +35,10 @@ static const char * progname;   // program name for error messages
 //--------------------------------------------------------------------------
 // Command line options flags
 static int TrimExif = FALSE;        // Cut off exif beyond interesting data.
-static int RenameToDate = FALSE;
+static int RenameToDate = 0;        // 1=rename, 2=rename all.
+#ifdef _WIN32
 static int RenameAssociatedFiles = FALSE;
+#endif
 static char * strftime_args = NULL; // Format for new file name.
 static int Exif2FileTime  = FALSE;
 static int DoModify     = FALSE;
@@ -86,7 +88,7 @@ static int ShowFileInfo = TRUE;     // Indicates to show standard file info
                                     // (file name, file size, file date)
 
 
-
+#define MATTHIAS
 #ifdef MATTHIAS
     // This #ifdef to take out less than elegant stuff for editing
     // the comments in a JPEG.  The programs rdjpgcom and wrjpgcom
@@ -145,7 +147,6 @@ static int FileEditComment(char * TempFileName, char * Comment, int CommentSize)
     {
         char * Editor;
         Editor = getenv("EDITOR");
-        if (strlen(Editor) > PATH_MAX) ErrFatal("env too long");
         if (Editor == NULL){
 #ifdef _WIN32
             Editor = "notepad";
@@ -153,11 +154,12 @@ static int FileEditComment(char * TempFileName, char * Comment, int CommentSize)
             Editor = "vi";
 #endif
         }
+        if (strlen(Editor) > PATH_MAX) ErrFatal("env too long");
 
         sprintf(QuotedPath, "%s \"%s\"",Editor, TempFileName);
         a = system(QuotedPath);
     }
-    
+
     if (a != 0){
         perror("Editor failed to launch");
         exit(-1);
@@ -294,38 +296,83 @@ static int AutoResizeCmdStuff(void)
 #endif // MATTHIAS
 
 
+//--------------------------------------------------------------------------
+// Escape an argument such that it is interpreted literally by the shell
+// (returns the number of written characters)
+//--------------------------------------------------------------------------
+static int shellescape(char* to, const char* from)
+{
+    int i, j;
+    i = j = 0;
+
+    // Enclosing characters in double quotes preserves the literal value of
+    // all characters within the quotes, with the exception of $, `, and \.
+    to[j++] = '"';
+    while(from[i])
+    {
+#ifdef _WIN32
+        // Under WIN32, there isn't really anything dangerous you can do with 
+        // escape characters, plus windows users aren't as sercurity paranoid.
+        // Hence, no need to do fancy escaping.
+        to[j++] = from[i++];
+#else
+        switch(from[i]) {
+            case '"':
+            case '$':
+            case '`':
+            case '\\':
+                to[j++] = '\\';
+                // Fallthru...
+            default:
+                to[j++] = from[i++];
+        }
+#endif 
+        if (j >= PATH_MAX) ErrFatal("max path exceeded");
+    }
+    to[j++] = '"';
+    return j;
+}
+
+
 //--------------------------------------------------------------------------
 // Apply the specified command to the JPEG file.
 //--------------------------------------------------------------------------
 static void DoCommand(const char * FileName, int ShowIt)
 {
     int a,e;
-    char ExecString[PATH_MAX*2];
-    char TempName[PATH_MAX+1];
+    char ExecString[PATH_MAX*3];
+    char TempName[PATH_MAX+10];
     int TempUsed = FALSE;
 
     e = 0;
 
-    // Make a temporary file in the destination directory by changing last char.
-    strcpy(TempName, FileName);
-    a = strlen(TempName)-1;
-    TempName[a] = (char)(TempName[a] == 't' ? 'z' : 't');
+    // Generate an unused temporary file name in the destination directory
+    // (a is the number of characters to copy from FileName)
+    a = strlen(FileName)-1;
+    while(a > 0 && FileName[a-1] != SLASH) a--;
+    memcpy(TempName, FileName, a);
+    strcpy(TempName+a, "XXXXXX");
+    mktemp(TempName);
+    if(!TempName[0]) {
+        ErrFatal("Cannot find available temporary file name");
+    }
+
+
 
     // Build the exec string.  &i and &o in the exec string get replaced by input and output files.
     for (a=0;;a++){
         if (ApplyCommand[a] == '&'){
             if (ApplyCommand[a+1] == 'i'){
                 // Input file.
-                e += sprintf(ExecString+e, "\"%s\"",FileName);
+                e += shellescape(ExecString+e, FileName);
                 a += 1;
                 continue;
             }
             if (ApplyCommand[a+1] == 'o'){
                 // Needs an output file distinct from the input file.
-                e += sprintf(ExecString+e, "\"%s\"",TempName);
+                e += shellescape(ExecString+e, TempName);
                 a += 1;
                 TempUsed = TRUE;
-                unlink(TempName);// Remove any pre-existing temp file
                 continue;
             }
         }
@@ -436,7 +483,7 @@ void RenameAssociated(const char * FileName, char * NewBaseName)
     FilePattern[ExtPos] = '*';
     FilePattern[ExtPos+1] = '\0';
 
-    for(PathLen = strlen(FileName);FileName[PathLen-1] != '\\';){
+    for(PathLen = strlen(FileName);FileName[PathLen-1] != SLASH;){
         if (--PathLen == 0) break;
     }
 
@@ -478,15 +525,17 @@ static void DoFileRenaming(const char * FileName)
 {
     int NumAlpha = 0;
     int NumDigit = 0;
-    int PrefixPart = 0;
-    int ExtensionPart = strlen(FileName);
+    int PrefixPart = 0; // Where the actual filename starts.
+    int ExtensionPart;  // Where the file extension starts.
     int a;
     struct tm tm;
     char NewBaseName[PATH_MAX*2];
     int AddLetter = 0;
+    char NewName[PATH_MAX+2];
 
+    ExtensionPart = strlen(FileName);
     for (a=0;FileName[a];a++){
-        if (FileName[a] == '/' || FileName[a] == '\\'){
+        if (FileName[a] == SLASH){
             // Don't count path component.
             NumAlpha = 0;
             NumDigit = 0;
@@ -513,7 +562,7 @@ static void DoFileRenaming(const char * FileName)
     }
     
 
-    strcpy(NewBaseName, FileName); // Get path component of name.
+    strncpy(NewBaseName, FileName, PATH_MAX); // Get path component of name.
 
     if (strftime_args){
         // Complicated scheme for flexibility.  Just pass the args to strftime.
@@ -566,14 +615,16 @@ static void DoFileRenaming(const char * FileName)
                 }
             }
         }
-
-        strftime(NewBaseName+PrefixPart, PATH_MAX, pattern, &tm);
+        strftime(NewName, PATH_MAX, pattern, &tm);
     }else{
         // My favourite scheme.
-        sprintf(NewBaseName+PrefixPart, "%02d%02d-%02d%02d%02d",
+        sprintf(NewName, "%02d%02d-%02d%02d%02d",
              tm.tm_mon+1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
     }
 
+    NewBaseName[PrefixPart] = 0;
+    CatPath(NewBaseName, NewName);
+
     AddLetter = isdigit(NewBaseName[strlen(NewBaseName)-1]);
     for (a=0;;a++){
         char NewName[PATH_MAX+10];
@@ -587,9 +638,9 @@ static void DoFileRenaming(const char * FileName)
             // is before the '.' in ascii, and so sorting the names would put the later name before
             // the name without suffix, causing the pictures to more likely be out of order.
             if (AddLetter){
-                NameExtra[0] = (char)('a'-1+a); // Try a,b,c,d... for suffix if it ends in a letter.
+                NameExtra[0] = (char)('a'-1+a); // Try a,b,c,d... for suffix if it ends in a number.
             }else{
-                NameExtra[0] = (char)('0'-1+a); // Try 1,2,3,4... for suffix if it ends in a char.
+                NameExtra[0] = (char)('0'-1+a); // Try 0,1,2,3... for suffix if it ends in a latter.
             }
             NameExtra[1] = 0;
         }else{
@@ -600,6 +651,11 @@ static void DoFileRenaming(const char * FileName)
 
         if (!strcmp(FileName, NewName)) break; // Skip if its already this name.
 
+        if (!EnsurePathExists(NewBaseName)){
+            break;
+        }
+
+
         if (stat(NewName, &dummy)){
             // This name does not pre-exist.
             if (rename(FileName, NewName) == 0){
@@ -614,6 +670,7 @@ static void DoFileRenaming(const char * FileName)
                 printf("Error: Couldn't rename '%s' to '%s'\n",FileName, NewName);
             }
             break;
+
         }
 
         if (a > 25 || (!AddLetter && a > 9)){
@@ -638,7 +695,7 @@ static int DoAutoRotate(const char * FileName)
                 ErrFatal("Orientation screwup");
             }
 
-            sprintf(RotateCommand, "jpegtran -%s -outfile &o &i", Argument);
+            sprintf(RotateCommand, "jpegtran -trim -%s -outfile &o &i", Argument);
             ApplyCommand = RotateCommand;
             DoCommand(FileName, FALSE);
             ApplyCommand = NULL;
@@ -657,7 +714,7 @@ static int DoAutoRotate(const char * FileName)
                 strcpy(ThumbTempName_out, FileName);
                 strcat(ThumbTempName_out, ".tho");
                 SaveThumbnail(ThumbTempName_in);
-                sprintf(RotateCommand,"jpegtran -%s -outfile \"%s\" \"%s\"",
+                sprintf(RotateCommand,"jpegtran -trim -%s -outfile \"%s\" \"%s\"",
                     Argument, ThumbTempName_out, ThumbTempName_in);
 
                 if (system(RotateCommand) == 0){
@@ -1143,7 +1200,7 @@ badtime:
 static void Usage (void)
 {
     printf("Jhead is a program for manipulating settings and thumnails in Exif jpeg headers\n"
-           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, Oct 4 2008.\n"
+           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, Feb 14 2009.\n"
            "http://www.sentex.net/~mwandel/jhead\n"
            "\n");
 
@@ -1188,6 +1245,8 @@ static void Usage (void)
            "             The '.jpg' is automatically added to the end of the name.  If the\n"
            "             destination name already exists, a letter or digit is added to \n"
            "             the end of the name to make it unique.\n"
+           "             The new name may include a path as part of the name.  If this path\n"
+           "             does not exist, it will be created\n"
            "  -nf[format-string]\n"
            "             Same as -n, but rename regardless of original name\n"
            "  -a         (Windows only) Rename files with same name but different extension\n"
@@ -1436,13 +1495,17 @@ int main (int argc, char **argv)
             if (*arg){
                 // A strftime format string is supplied.
                 strftime_args = arg;
+                #ifdef _WIN32
+                    SlashToNative(strftime_args);
+                #endif
                 //printf("strftime_args = %s\n",arg);
             }
         }else if (!strcmp(arg,"-a")){
-            RenameAssociatedFiles = TRUE;
             #ifndef _WIN32
                 ErrFatal("Error: -a only supported in Windows version");
-            #endif 
+            #else
+                RenameAssociatedFiles = TRUE;
+            #endif
         }else if (!strcmp(arg,"-ft")){
             Exif2FileTime = TRUE;
             DoReadAction = TRUE;
@@ -1602,13 +1665,7 @@ int main (int argc, char **argv)
         FilesMatched = FALSE;
 
         #ifdef _WIN32
-            {
-                int a;
-                for (a=0;;a++){
-                    if (argv[argn][a] == '\0') break;
-                    if (argv[argn][a] == '/') argv[argn][a] = '\\';
-                }
-            }
+            SlashToNative(argv[argn]);
             // Use my globbing module to do fancier wildcard expansion with recursive
             // subdirectories under Windows.
             MyGlob(argv[argn], ProcessFile);
